/*
 * Copyright (C) 2006-2007 Funambol
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

package com.funambol.mail;

import com.funambol.util.StringUtil;
import java.util.Hashtable;
import java.util.Random;
import java.util.Date;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import com.funambol.storage.Serializable;
import com.funambol.storage.ComplexSerializer;
import com.funambol.util.Log;

/**
 * This class models a MIME-style email message. <p>
 *
 * <code>Message</code> extends the <code>Part</code> abstract class, and
 * contains a set of attributes and a 'content'. Messages within a folder also
 * have a set of flags that describe its state within the folder. This state is
 * in turn described by the calss MessageFlags. <p>
 *
 * <code>Message</code> defines some new attributes in addition to those
 * defined in <code>Part</code>. These attributes specify meta-data for the
 * message -i.e., addressing and descriptive information about the message. <p>
 *
 * Clients create new MIME-style messages building an empty <code>Message</code>
 * object and then filling it with appropriate attributes and content.
 * <code>Message</code> objects can be obtained from a <code>Folder</code>
 * too. <p>
 *
 * TODO: A Message object obtained from a folder is just a lightweight reference
 * to the actual message. The Message is 'lazily' filled up (on demand) when
 * each item is requested from the message. Note that certain folder
 * implementations may return Message objects that are pre-filled with certain
 * user-specified items. <p>
 *
 */
public class Message extends Part implements Serializable {
    
    // -------------------------------------------------------------- Constants
    
    /** TO header name */
    public static final String TO = "To";
    
    /** FROM header name */
    public static final String FROM = "From";
    
    /** CC header name */
    public static final String CC = "Cc";
    
    /** BCC header name */
    public static final String BCC = "Bcc";
    
    /** REPLY-TO header name */
    public static final String REPLYTO = "Reply-To";
    
    /** SUBJECT header name */
    public static final String SUBJECT = "Subject";
    
    // ------------------------------------------------------------- Attributes
    
    /**
     * The unique ID of this message.
     *
     * This is an string containing the global-unique message ID. It is
     * generated by the application for outgoing messages and is taken from the
     * received message otherewise.
     */
    protected String msgid;
    
    /**
     * A combination of status flags for this <code>Message</code> within a
     * folder, e.g. 'read', 'forwarded', 'draft' etc.
     */
    protected MessageFlags flags;
    
    /**
     * A Folder can contain many Messages, and each Message has a reference to
     * its Folder
     *
     */
    protected Folder parent;
    
    /**
     * The sent date of this message
     */
    protected Date sent;
    
    /**
     * The received date of this message
     */
    protected Date received;
    
    /**
     * The content of this Message as an Object (it can be a Multipart or a
     * String)
     *
     * When it is a String, it is stored in the native encoding of the
     * application, regardless of how it has been transferred. During parsing
     * operation, the String is decoded according to the
     * 'Content-Transfer-Encoding' field.
     */
    protected Object content;
    
    /**
     * True is the message has been read by the store but without content
     */
    protected boolean incomplete;
    
    
    // ----------------------------------------------------------- Constructors
    
    /**
     * The default constructor <p>
     *
     * Initialize the message with default values suitable for sending it
     * (content-type: text/plain, sent and received date equal to current time).
     * The caller must override these values with the desired content.
     */
    public Message() {
        super(); // init Part
        
        msgid = createUniqueMessageIDValue();
        flags = new MessageFlags();
        parent = null;
        sent = new Date();
        received = sent;
        content = null;
        incomplete = false;
    }
    
    // --------------------------------------------------------- Public Methods
    
    /**
     * Add the specified recipients to the correct header, based on the Address
     * type.
     *
     * @param list
     *            list of Adresses to add to the recipient header.
     */
    public void addRecipients(Address[] list) throws MailException {
        int length = list.length;
        String name = null;
        
        for (int i = 0; i < length; i++) {
            if(list[i] != null) {
                name = typeToName(list[i].getType());
                if (name != null) {
                    addRecipient(name, list[i].toString());
                }
            }
        }
    }
    
    /**
     * Add the specified recipient to the correct header, based on the Address
     * type (see com.funambol.mail.Address for the defined types).
     *
     * @param address
     *            the address to add to the recipient header.
     */
    public void addRecipient(Address address) throws MailException {
        String name=null;
        name = typeToName(address.getType());
        if (name != null && address != null) {
            addRecipient(name, address.toString());
        }
    }
    
    /**
     * Get the sender address
     */
    public Address getFrom() throws MailException {
        String header = getHeader(FROM);
        if ((header!=null) && (!header.equals(""))) {
            return new Address(Address.FROM, header);
        } else {
            return null;
        }
    }
    
    
    /**
     * Get the destination recipients
     */
    public Address[] getTo() throws MailException {
        Address[] addrlist = Address.parse(Address.TO, getHeader(TO));
        
        return addrlist;
    }
    
    
    /**
     * Get the Cc: recipients
     */
    public Address[] getCc() throws MailException {
        Address[] addrlist = Address.parse(Address.CC, getHeader(CC));
        
        return addrlist;
    }
    
    
    /**
     * Get the Bcc: recipients
     */
    public Address[] getBcc() throws MailException {
        return Address.parse(Address.BCC, getHeader(BCC));
    }
    
    
    /**
     * Get the Reply-To: address (to from if not specified)
     */
    public Address[] getReplyTo() throws MailException {
        String replyto = getHeader(REPLYTO);
        if (StringUtil.isNullOrEmpty(replyto)) {
            return Address.parse(Address.TO, getHeader(FROM));
        } else {
            return Address.parse(Address.TO, replyto);
        }
    }
    
    
    /**
     * Returns the flag mask for this message.
     *
     * @return the flag mask for this message.
     */
    public MessageFlags getFlags() {
        return flags;
    }
    
    
    /**
     * Returns the String representing the ID of this Message.
     *
     * @return The unique ID of this Message
     */
    public String getMessageId() {
        return msgid;
    }
    
    
    /**
     * Convenience method to return the content of the RFC 2822 "Subject:"
     * header field
     */
    public String getSubject() {
        return getHeader(SUBJECT);
    }
    
    
    /**
     * Returns the sent date of this message.
     *
     * @return the sent date of this message.
     */
    public Date getSentDate() {
        return sent;
    }
    
    
    /**
     * Sets the flag mask for this message.
     *
     * @param flags the flag mask for this message.
     */
    public void setFlags(MessageFlags flags) {
        this.flags.setFlags(flags.getFlags());
    }
    
    
    /**
     * Returns the received date of this message.
     *
     * @return the received date of this message.
     */
    public Date getReceivedDate() {
        return received;
    }
    
    
    /**
     * This method removes all the recipients of the specified type from this
     * message.
     *
     * @param type
     *            one of the defined types in
     *
     */
    public void removeAllRecipients(int type) {
        try {
            String name = typeToName(type);
            removeHeader(name);
        } catch (MailException me) {
            // Don't propagate exception for remove.
            Log.error(this, "removeAllRecipients() exception raised by " +
                    "typeToname for type " + type);
            me.printStackTrace();
        }
    }
    
    /**
     * This method sets the content of the RFC 2822 "From:" header with the
     * given <code>Address</code>
     */
    public void setFrom(Address from) {
        if (from==null) {
            setHeader(FROM, "");
        } else {
            setHeader(FROM, from.toString());
        }
    }
    
    
    /**
     * Set the content of the RFC 2822 "To:" header with the provided addresses,
     * passed as an Address array
     *
     * @param tolist
     *            list of Adresses to put in the To: list
     */
    public void setTo(Address[] tolist) {
        setRecipients(TO, tolist);
    }
    
    
    /**
     * Set the content of the RFC 2822 "Cc:" (carbon copy) header with the
     * provided addresses, passed as an Address array
     *
     * @param cclist
     *            list of Adresses to put in the Cc: list
     */
    public void setCc(Address[] cclist) {
        setRecipients(CC, cclist);
    }
    
    
    /**
     * Set the content of the RFC 2822 "Bcc:" (blind carbon copy) header with
     * the provided addresses, passed as an Address array
     *
     * @param bcclist
     *            list of Adresses to put in the bcclist
     */
    public void setBcc(Address[] bcclist) {
        setRecipients(BCC, bcclist);
    }
    
    
    /**
     * Set the content of the RFC 2822 "Bcc:" (blind carbon copy) header with
     * the provided addresses, passed as an Address array
     *
     * @param replytolist
     *            list of Adresses to put in the bcclist
     */
    public void setReplyTo(Address[] replytolist) {
        setRecipients(REPLYTO, replytolist);
    }
    
    
    /**
     * Set the content of the RFC 2822 "Subject:" header.
     *
     * @param subject
     *            The text to put in the Subject
     */
    public void setSubject(String subject) {
        setHeader(SUBJECT, subject);
    }
    
    
    /**
     * Returns this message's parent Folder
     *
     * @return A reference to the Folder object this Message is contained in
     */
    public Folder getParent() {
        return parent;
    }
    
    
    /**
     * Sets a reference to the <code>Folder</code> this <code>Message</code>
     * is contained in
     *
     * @param parent
     *            The folder this <code>Message</code> is a child of
     */
    public void setParent(Folder parent) {
        this.parent = parent;
    }
    
    
    /**
     * Set the sent date
     *
     * @param sent
     *            the sent date
     */
    public void setSentDate(Date sent) {
        this.sent = sent;
    }
    
    
    /**
     * Set the received date
     *
     * @param received
     *            the received date
     */
    public void setReceivedDate(Date received) {
        this.received = received;
    }
    
    
    /**
     * Get the content of this message: can be a String or a Multipart.
     *
     */
    public Object getContent() {
        return content;
    }
    
    
    /**
     * Convenience method to get the body for text messages
     *
     * TODO: retrieve the text of the first BodyPart for MultiPart?
     *
     * @return the text content of the message, or null this message has no text
     *         content.
     */
    public String getTextContent() {
        if (content instanceof String) {
            return (String)content;
        } else {
            return null;
        }
    }
    
    
    /**
     * Sets the content of this Message. It can be a Multipart or a String. This
     * method sets also the contentType: multipart/mixed if the content is
     * Multipart, text/plain if it's a String
     *
     */
    public void setContent(Object content) throws MailException {
        if (content instanceof Multipart) {
            ((Multipart)content).setContainer(this);
            contentType = "multipart/mixed";
        } else if (content instanceof String) {
            contentType = "text/plain";
        } else {
            throw (new MailException(
                        MailException.INVALID_CONTENT,
                        "Invalid content: " + content.getClass().getName()));
        }
        
        //see explication in Part.setContentType
        setContentType(contentType);
        this.content = content;
    }
    
    
    /**
     * Sets the content of this Message. This can be another Multipart (in case
     * of a multi-content message) or a String (in case of a 'single-content'
     * message)
     *
     * @param content
     *            The content object
     * @param type
     *            The MIME type of the content
     */
    public void setContent(Object content, String type) throws MailException {
        setContent(content);
        setContentType(type);
    }
    
    
    /**
     * Sets the String representing the unique ID of this Message
     */
    public void setMessageId(String messageId) {
        msgid = messageId;
    }
    
    
    // ---------------------------- Serializable implementation
    
    public void serialize(DataOutputStream dout) throws IOException {
        // This message has been retrieved incomplete from the store
        if (incomplete) {
            throw new IOException("Can't serialize an incomplete Message");
        }
        
        dout.writeUTF(msgid);
        flags.serialize(dout);
        
        // Parent must be set by Folder when the message is retrieved,
        // so it's not saved.
        
        // Write Part attributes
        ComplexSerializer.serializeHashTable(dout, headers);
        dout.writeUTF(contentType);
        
        dout.writeLong(sent.getTime());
        dout.writeLong(received.getTime());
        
        ComplexSerializer.serializeObject(dout, content);
        
    }
    
    
    public void deserialize(DataInputStream din) throws IOException {
        msgid = din.readUTF();
        // Deserialize flags
        flags.deserialize(din);
        
        headers = ComplexSerializer.deserializeHashTable(din);
        
        contentType = din.readUTF();
        sent = new Date(din.readLong());
        received = new Date(din.readLong());
        
        content = ComplexSerializer.deserializeObject(din);
        incomplete = false;
        
        // Set the container reference for Multipart
        if (content instanceof Multipart) {
            ((Multipart)content).setContainer(this);
        }
        
    }
    
    
    // -------------------------------------------------------- Private Methods
    
    /**
     * Converts the address type in a header name
     *
     * @param type the type if the address (see com.funambol.mail.Address )
     * @return the header name
     */
    private String typeToName(int type) throws MailException {
        switch (type) {
            case Address.FROM:
                return Message.FROM;
            case Address.TO:
                return Message.TO;
            case Address.CC:
                return Message.CC;
            case Address.BCC:
                return Message.BCC;
            case Address.REPLYTO:
                return Message.REPLYTO;
            default:
                throw new MailException(
                            MailException.INVALID_ADDRESS,
                            "Invalid address type: "+type);
        }
    }
    
    
    /**
     * @param name
     * @param value
     */
    private void addRecipient(String name, String value) {
        String header = getHeader(name);
        
        
        if (header != null) {
            if (header.trim().length()!=0) {
                header = new StringBuffer(header).append(", ")
                .append(value).toString();
            } else {
                header = value;
            }
            
        } else {
            header = value;
        }
        setHeader(name, header);
    }
    
    
    /**
     * Set the specified recipient with the provided addresses, passed as an
     * Address array
     *
     * @param recipient
     *            the name of the recipient to set
     * @param list
     *            list of Adresses to put in the bcclist
     */
    private void setRecipients(String recipient, Address[] list) {
        int length = list.length;
        if (length>0) {
            StringBuffer buffer = new StringBuffer(list[0].toString());
            
            for (int i = 1; i < length; i++) {
                if(list[i] != null && !"".equals(list[i])) {
                    buffer.append(", ").append(list[i].toString());
                }
            }
            
            setHeader(recipient, buffer.toString());
        }
    }
    
    
    /**
     * Creates a unique value for use in the Message ID <p>
     *
     * This implementation generates it by concatenating a newly created
     * object's <code>hashCode()</code>, a random long with seed from the the
     * current time, and the string "@Funambol.mail".
     *
     * TODO: usually the message ID ends with the domain name of the host
     * originating the message, or the one of the user's mail address, but this
     * is not easy to do on the phone.
     */
    public String createUniqueMessageIDValue() {
        Random r = new Random();
        StringBuffer s = new StringBuffer();
        
        // Unique string is <timestamp><random>@Funambol.mail
        s.append(System.currentTimeMillis()).append(r.nextInt()).append(
                "@Funambol.mail");
        
        return s.toString();
    }
    
}
